#include "kitutuki.h"
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <locale.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <glob.h>
#include <libgen.h>
#include <dirent.h>
#include <oniguruma.h>
#include <wctype.h>
#include <wchar.h>

#include "config.h"

#if defined(HAVE_CURSES_H)
#include <curses.h>
#elif defined(HAVE_NCURSES_H)
#include <ncurses.h>
#elif defined(HAVE_NCURSES_NCURSES_H)
#include <ncurses/ncurses.h>
#endif

#ifdef __LINUX__
#include <sys/stat.h>
#include <linux/stat.h>
#endif
#if defined(__CYGWIN__) || defined(__CYGWIN7__)
#include <sys/stat.h>
#endif
#
#ifdef __LINUX__
#include <linux/limits.h>
#endif

#ifdef __LINUX__

#include <linux/limits.h>
#define __USE_POSIX
#define __USE_XOPEN_EXTENDED
#define __USE_POSIX199309
#define __USE_UNIX98
#include <signal.h>
#include <sys/wait.h>

#else

#include <signal.h>
#include <sys/stat.h>

#endif

#include "minato_extra.h"

void srandom(unsigned int seed);
char *strdup(const char *s);
int setenv(const char *name, const char *value, int overwrite);
int unsetenv(const char *name);
int strcasecmp(const char *s1, const char *s2);
int wcwidth(wchar_t c);
int wcswidth(const wchar_t *s, size_t n);
size_t wcslen(const wchar_t *s);

#define PATH_MAX 4096

string_obj* gPrompt;
string_obj* gCmdLine;

vector_obj* gHistory;
int gHistoryCursor;
int gCmdLineCursor;    // Хñ̡UTF8ΰʸñ̤ǤϤʤ

BOOL gCmdLineLoop = FALSE;

void cmdline_init()
{
    gPrompt = STRING_NEW("");
    gHistory = VECTOR_NEW(50);

    gHistoryCursor = 0;
    gCmdLineCursor = 0;

    gCmdLine = STRING_NEW("");
}

void cmdline_final()
{
    int i;

    string_delete(gPrompt);

    for(i=0; i<vector_size(gHistory); i++) {
        string_delete(vector_item(gHistory, i));
    }
    vector_delete(gHistory);

    string_delete(gCmdLine);
}

static void view()
{
}

/// ɽѤѿ ///
int gKanjiZure[128];   // źϹԡιԤˤ줬äɤ
int gCmdLineCursorX;
int gCmdLineCursorY;

static BOOL is_utf8(char* mbs)
{
}

static BOOL kanji_continue(char* mbs)
{
    wchar_t* wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*((strlen(mbs)+1)*MB_CUR_MAX));

    if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) < 0) {
        perror("mbstowcs");
        exit(1);
    }

    BOOL result = strlen(mbs) == str_kanjipos2pointer(kUtf8, mbs, wcslen(wcs)-1) -mbs;

    FREE(wcs);

    return result;
}


static void get_terminal_stat(int prompt_termlen, int* x, int* line, int* cursor_x, int* cursor_line, BOOL* kanji_zure, BOOL* new_line, BOOL* fukanzen)
{
    const int maxx = mgetmaxx();
    const int maxy = mgetmaxy();

    char* mbs = string_c_str(gCmdLine);
    wchar_t* wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*((strlen(mbs)+1)*MB_CUR_MAX));

    if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) < 0) {
        perror("mbstowcs");
        exit(1);
    }

    /// Ϥʤ ///
    if(strlen(mbs) == 0 || wcslen(wcs) == 0) {
        *x = prompt_termlen;
        *line = 0;
        *new_line = FALSE;
        *cursor_x = prompt_termlen;
        *cursor_line = 0;
    }
    else {
        /// Ϥ ///
        if(is_utf8_bytes(mbs)) {
            int utf_cursor = str_pointer2kanjipos(kUtf8, mbs, mbs + gCmdLineCursor);

            *fukanzen = FALSE;

            int i;
            *line = 0;
            *x = prompt_termlen;
            *cursor_x = 0;
            *cursor_line = 0;
            for(i=0; i<wcslen(wcs); i++) {
                *new_line = FALSE;

                if(i == utf_cursor) {
                    *cursor_x = *x;
                    *cursor_line = *line;
                }

                (*x) += wcwidth(wcs[i]);

                if(*x > maxx) {
                    kanji_zure[(*line)++] = TRUE;
                    *x = *x - maxx + 1;
                    *new_line = TRUE;
                }
                else if(*x == maxx) {
                    kanji_zure[(*line)++] = FALSE;
                    *x = *x - maxx;
                    *new_line = TRUE;
                }
            }

            if(i == utf_cursor) {
                *cursor_x = *x;
                *cursor_line = *line;

            }
        }
        /// Դ ///
        else {
            *fukanzen = TRUE;
            char* mbs2 = MALLOC(strlen(mbs)+1);

            int i;
            for(i=1; i<MB_CUR_MAX; i++) {
                memcpy(mbs2, mbs, strlen(mbs)-i);
                mbs2[strlen(mbs)-i] = 0;
                if(is_utf8_bytes(mbs2)) {
                    break;
                }
            }

            const int fukanzen_len = i;

            int utf_cursor = str_pointer2kanjipos(kUtf8, mbs2, mbs2 + gCmdLineCursor-i);
            if(mbstowcs(wcs, mbs2, (strlen(mbs)+1)*MB_CUR_MAX) < 0) {
                perror("mbstowcs");
                exit(1);
            }

            FREE(mbs2);

            *line = 0;
            *x = prompt_termlen;
            *cursor_x = 0;
            *cursor_line = 0;
            for(i=0; i<wcslen(wcs); i++) {
                if(i == utf_cursor) {
                    *cursor_x = *x;
                    *cursor_line = *line;

                }
                *new_line = FALSE;
                (*x) += wcwidth(wcs[i]);

                if(*x > maxx) {
                    kanji_zure[(*line)++] = TRUE;
                    *x = *x - maxx + 1;
                    *new_line = TRUE;
                }
                else if(*x == maxx) {
                    kanji_zure[(*line)++] = FALSE;
                    *x = *x - maxx;
                    *new_line = TRUE;
                }
            }

            if(i == utf_cursor) {
                *cursor_x = *x;
                *cursor_line = *line;

            }

            *new_line = FALSE;
            (*x) += fukanzen_len;

            if(*x > maxx) {
                kanji_zure[(*line)++] = TRUE;
                *x = *x - maxx + 1;
                *new_line = TRUE;
            }
            else if(*x == maxx) {
                kanji_zure[(*line)++] = FALSE;
                *x = *x - maxx;
                *new_line = TRUE;
            }
        }
    }

    FREE(wcs);
}

/// ư ///
static void cursor_move(int v)
{
    char* str = string_c_str(gCmdLine);
    int utfpos = str_pointer2kanjipos(kUtf8, str, str + gCmdLineCursor);
    utfpos+=v;
    gCmdLineCursor = str_kanjipos2pointer(kUtf8, str, utfpos) - str;
}

static void input(int prompt_termlen)
{
    const int maxx = mgetmaxx();
    const int maxy = mgetmaxy();

    /// input ///
    int meta;
    int key = mgetch(&meta);

    int x;
    int line;
    BOOL kanji_zure[128];
    BOOL new_line;
    BOOL fukanzen;
    int cursor_x;
    int cursor_line;
    get_terminal_stat(prompt_termlen, &x, &line, &cursor_x, &cursor_line, kanji_zure, &new_line, &fukanzen);

    /// ϤϴλƤ ///
    if(key == 10 || key == 13) {       // CTRL-J CTRL-M
        gCmdLineLoop = FALSE;
    }
    /// ˰ư ///
    else if(key == 2 || key == KEY_LEFT) {    // CTRL-B
        /// Υư ///
        if(gCmdLineCursor > 0) {
            char* str = string_c_str(gCmdLine);

            int utfpos = str_pointer2kanjipos(kUtf8, str, str + gCmdLineCursor);
            gCmdLineCursor = str_pointer2kanjipos(kUtf8, str, str + gCmdLineCursor-1);
            int n = str_termlen_range(kUtf8, str, gCmdLineCursor, utfpos);

            /// ̾Υư ///
            if(cursor_line == 0 && cursor_x > prompt_termlen) {
                mcursor_left_immediately(n);
            }
            else if(cursor_x == 0) {
            }
            else {
                mcursor_left_immediately(n);
            }
        }
    }
    /// ˰ư ///
    else if(key == 6 || key == KEY_RIGHT) {    // CTRL-F
        /// Υư ///
        cursor_move(1);

        /// ̾Υư ///
        if(x == maxx) {
        }
        else if(x) {
            mcursor_right_immediately(1);
        }
    }
    /// ʸ ///
    else if(key == 8 || key == KEY_BACKSPACE) {    // CTRL-H
        if(gCmdLineCursor > 0) {
            string_obj* str = gCmdLine;
            char* cstr = string_c_str(str);

            /// ʸ ///
            const int utfpos = str_pointer2kanjipos(kUtf8, cstr, cstr + gCmdLineCursor);
            char* before_point = str_kanjipos2pointer(kUtf8, cstr, utfpos-1);
            const int erase_len = cstr + gCmdLineCursor - before_point;
            const int new_cursor = before_point-cstr;

            char buf[128];
            memcpy(buf, before_point, erase_len);
            buf[erase_len] = 0;

            int n = str_termlen(kUtf8, buf);

            string_erase(str, before_point - cstr, erase_len);

            /// Υ֤򹹿 ///
            gCmdLineCursor = new_cursor;

            /// ̾ʸ ///
            if(x==2 && kanji_zure[line-1]) {
                mbackspace_immediately();
                mbackspace_immediately();
                mbackspace_head_of_line_immediately();
            }
            else if(x==0) {
                mbackspace_head_of_line_immediately();
                int i;
                for(i=1; i<n; i++) {
                    mbackspace_immediately();
                }
            }
            else {
                int i;
                for(i=0; i<n; i++) {
                    mbackspace_immediately();
                }
            }
        }
    }
    else {
        /// ʸɲ ///
        if(meta == 0) {
            /// ʸϤ߽Фϲ ///
            /// (ʸUTF8ʸ2Ȳ) ///
            if(!fukanzen && maxx - x == 1 && (key < ' ' || key > '~')) {
                puts("");
                mprintw_immediately("%c", key);
            }
            else {
                mprintw_immediately("%c", key);
            }

            /// Υޥɥ饤򹹿 ///
            char tmp[128];
            sprintf(tmp, "%c", key);
            string_insert(gCmdLine, gCmdLineCursor, tmp);

            gCmdLineCursor++;

            /// ɽ奫뤬ǸƤ ///
            int x;
            int line;
            BOOL kanji_zure[128];
            BOOL new_line;
            BOOL fukanzen;
            int cursor_x;
            int cursor_line;
            get_terminal_stat(prompt_termlen, &x, &line, &cursor_x, &cursor_line, kanji_zure, &new_line, &fukanzen);

            if(x == 0 && !fukanzen) {
                puts("");
            }
        }
    }
}

char* mreadline(char* prompt)
{
    string_put(gPrompt, prompt);
    string_put(gCmdLine, "");
    gCmdLineCursor = 0;

    memset(gKanjiZure, 0, sizeof(int)*128); // ɽѤΥ
    gCmdLineCursorX = 0;
    gCmdLineCursorY = 0;;

    memset(gKanjiZure, 0, sizeof(int)*128);

    minitscr();
    mmove_line_home_immediately();
    // Τ׻뤿˥ץץȤɬƬϤ᤿
    mprintw_immediately("%s", string_c_str(gPrompt));
    int prompt_termlen = str_termlen(kUtf8, string_c_str(gPrompt));

    gCmdLineLoop = TRUE;
    while(gCmdLineLoop) {
        input(prompt_termlen);
    }

    mendwin();

    puts("");

    return string_c_str(gCmdLine);
}

void madd_history(char* cmdline)
{
}

void mwrite_history(char* file)
{
}

/// ͤɤ߹ҥȥο
int mread_history(char* file)
{
}

/// numƬ0
void mremove_history(int num)
{
}

////////////////////////////////////////////////////////////
// եǽ
////////////////////////////////////////////////////////////
/*
typedef struct {
    string_obj* mName;

    struct stat mStat;                // stat()Υå
} sFile;

sFile* sFile_new(char* name, struct stat* stat_)
{
    sFile* self = MALLOC(sizeof(sFile));

    self->mName = STRING_NEW(name);
    self->mStat = *stat_;

    return self;
}

void sFile_delete(sFile* self)
{
    string_delete(self->mName);
    FREE(self);
}

static BOOL sort_name(void* left, void* right)
{
    sFile* l = (sFile*) left;
    sFile* r = (sFile*) right;

    char* lfname = string_c_str(l->mName);
    char* rfname = string_c_str(r->mName);

    if(strcmp(lfname, ".") == 0) return TRUE;
    if(strcmp(lfname, "..") == 0) {
        if(strcmp(rfname, ".") == 0) return FALSE;

        return TRUE;          
    }
    if(strcmp(rfname, ".") == 0) return FALSE;
    if(strcmp(rfname, "..") == 0) return FALSE;

    if(!S_ISDIR(l->mStat.st_mode) && !S_ISDIR(r->mStat.st_mode)
        || S_ISDIR(l->mStat.st_mode) && S_ISDIR(r->mStat.st_mode) )
    {
        return strcasecmp(lfname, rfname) < 0;
    }

    if(S_ISDIR(l->mStat.st_mode))
        return TRUE;
    else
        return FALSE;
}


// pathκǸ/Ĥ
static BOOL readline_select_file_read(vector_obj* v, char* path)
{
    DIR* dir = opendir(path);
    if(dir == NULL) {
        return FALSE;
    }
    
    struct dirent* entry;
    while(entry = readdir(dir)) {
        char fpath[PATH_MAX];
        strcpy(fpath, path);
        strcat(fpath, entry->d_name);

        struct stat stat_;
        memset(&stat_, 0, sizeof(struct stat));
        if(stat(fpath, &stat_) < 0) {
            continue;
        }

        sFile* file = sFile_new(entry->d_name, &stat_);

        vector_add(v, file);
    }
    
    closedir(dir);

    ///  ///
    vector_sort(v, sort_name);

    return TRUE;
}

static int readline_select_file(int count, int key)
{
    //msave_screen();
    minitscr();

    const int maxx = mgetmaxx();
    const int maxy = mgetmaxy();

    vector_obj* v = VECTOR_NEW(10);
    char cwd_path[PATH_MAX];
    mygetcwd(cwd_path);
    strcat(cwd_path, "/");

    char current_dir[PATH_MAX];
    strcpy(current_dir, cwd_path);
    readline_select_file_read(v, current_dir);

    vector_obj* mark_files = VECTOR_NEW(10);

    if(vector_size(v) > 0) {
        int cursor = 0;
        while(1) {
            ///  ///
            mclear();
            mmvprintw(0,0,"%s", current_dir);

            int page = cursor/((maxy-1)*4);
            int n = page*4*(maxy-1);
            while(n < page*4*(maxy-1)+(maxy-1)*4 && n < vector_size(v)) {
                sFile* file = vector_item(v, n);

                int attrs = 0;
                if(n == cursor) {
                    attrs |= kCAReverse;
                }
                if(S_ISDIR(file->mStat.st_mode)) {
                    attrs |= kCACyan;
                    attrs |= kCABold;
                }
                if(attrs) mattron(attrs);

                char file2[PATH_MAX];
                str_cut2(kUtf8, string_c_str(file->mName)
                        , maxx/4-2, file2, PATH_MAX);

                int y = (n-page*4*(maxy-1)) / 4 + 1;
                int x = ((n-page*4*(maxy-1)) % 4)*maxx/4;

                char full_path[PATH_MAX];
                strcpy(full_path, current_dir);
                strcat(full_path, string_c_str(file->mName));

                BOOL marked = FALSE;
                int i;
                for(i=0; i<vector_size(mark_files); i++) {
                    string_obj* mark_file = vector_item(mark_files, i);

                    if(strcmp(string_c_str(mark_file), full_path) == 0) {
                        marked = TRUE;
                        break;
                    }
                }

                if(marked) 
                    mmvprintw(y, x, "*%s", file2);
                else
                    mmvprintw(y, x, " %s", file2);

                if(attrs) {
                    mattroff();
                }

                n++;
            }
            mrefresh();

            /// ץå ///
            int meta;
            int key = mgetch(&meta);

            if(key == 14 || key == KEY_DOWN) {
                cursor+=4;
            }
            else if(key == 16 || key == KEY_UP) {
                cursor-=4;
            }
            else if(key == 6 || key == KEY_RIGHT) {
                cursor++;
            }
            else if(key == 2 || key == KEY_LEFT) {
                cursor--;
            }
            else if(key == 4 || key == KEY_NPAGE) {
                cursor += (maxy-1)*4;
            }
            else if(key == 21 || key == KEY_PPAGE) {
                cursor -= (maxy-1)*4;
            }
            else if(key == 12) {
                mclear();
                mrefresh();
            }
            else if(key == ' ') {
                sFile* cursor_file = vector_item(v, cursor);

                string_obj* cursor_mfile = NULL;

                char cursor_path[PATH_MAX];
                strcpy(cursor_path, current_dir);
                strcat(cursor_path, string_c_str(cursor_file->mName));

                int i;
                for(i=0; i<vector_size(mark_files); i++) {
                    string_obj* mfile = vector_item(mark_files, i);

                    if(strcmp(string_c_str(mfile), cursor_path) == 0) {
                        cursor_mfile = mfile;
                        break;
                    }
                }

                if(!cursor_mfile) {
                    vector_add(mark_files, STRING_NEW(cursor_path));
                }
                else {
                    vector_erase(mark_files
                        , vector_index(mark_files, cursor_mfile));
                    string_delete(cursor_mfile);
                }

                cursor++;
            }
            else if(key == '\\') {
                strcpy(current_dir, "/");

                int i;
                for(i=0; i<vector_size(v); i++) {
                    sFile_delete(vector_item(v, i));
                }
                vector_clear(v);

                readline_select_file_read(v, current_dir);
                cursor = 0;
            }
            else if(key == '\t' || key == 'w') {
                if(vector_size(mark_files) == 0) {
                    sFile* cursor_file = vector_item(v, cursor);

                    /// ߤΥȥǥ쥯ȥΥѥʤ
                    /// ǥ쥯ȥ̾
                    char cursor_path[PATH_MAX];
                    strcpy(cursor_path, current_dir);
                    strcat(cursor_path, string_c_str(cursor_file->mName));

                    /// ȤΥե롢ҥǥ쥯ȥΥեʤ
                    /// Хǥ쥯ȥ
                    if(memcmp(cursor_path
                            , cwd_path, strlen(cwd_path)) == 0 )
                    {
                        char cursor_path2[PATH_MAX];
                        memcpy(cursor_path2
                            , cursor_path+strlen(cwd_path)
                            , strlen(cursor_path) - strlen(cwd_path)+1);

                        string_obj* quoted_fname = STRING_NEW("");
                        kitutuki_get_quoted_fname(cursor_path2
                                        , quoted_fname);
                        rl_insert_text(string_c_str(quoted_fname));
                        string_delete(quoted_fname);
                    }
                    else {
                        string_obj* quoted_fname = STRING_NEW("");
                        kitutuki_get_quoted_fname(cursor_path, quoted_fname);
                        rl_insert_text(string_c_str(quoted_fname));
                        string_delete(quoted_fname);
                    }
                }
                else {
                    int j;
                    for(j=0; j<vector_size(mark_files); j++) {
                        char full_path[PATH_MAX];
                        strcpy(full_path
                            , string_c_str(vector_item(mark_files, j)));

                        if(strstr(full_path, cwd_path) == full_path) {
                            char full_path2[PATH_MAX];
                            memcpy(full_path2
                                , full_path + strlen(cwd_path) + 1
                                , strlen(full_path) - strlen(cwd_path));

                            string_obj* quoted_fname = STRING_NEW("");
                            kitutuki_get_quoted_fname(full_path2
                                            , quoted_fname);
                            rl_insert_text(string_c_str(quoted_fname));
                            string_delete(quoted_fname);

                            if(j<vector_size(mark_files)-1) 
                                rl_insert_text(" ");
                        }
                        else {
                            string_obj* quoted_fname = STRING_NEW("");
                            kitutuki_get_quoted_fname(full_path, quoted_fname);
                            rl_insert_text(string_c_str(quoted_fname));
                            string_delete(quoted_fname);

                            if(j<vector_size(mark_files)-1) 
                                rl_insert_text(" ");
                        }
                    }
                }
                break;
            }
            else if(key == 'q' || key == 3 | key == 27 || key == 7) {
                break;
            }
            else if(key == 10 || key == 13) {
                sFile* cursor_file = vector_item(v, cursor);

                if(strcmp(string_c_str(cursor_file->mName), ".") != 0
                    && S_ISDIR(cursor_file->mStat.st_mode)) 
                {
                    char path2[PATH_MAX];
                    if(correct_path(current_dir
                            , string_c_str(cursor_file->mName), path2)) 
                    {
                        strcpy(current_dir, path2);
                        if(current_dir[strlen(current_dir)-1] != '/') {
                            strcat(current_dir, "/");
                        }

                        int i;
                        for(i=0; i<vector_size(v); i++) {
                            sFile_delete(vector_item(v, i));
                        }
                        vector_clear(v);

                        readline_select_file_read(v, current_dir);
                        cursor = 0;
                    }
                }
            }
            else if(key == 8 || key == KEY_BACKSPACE) {    // CTRL-H
                char parent_path[PATH_MAX];
                parentname(parent_path, current_dir);
                strcpy(current_dir, parent_path);

                int i;
                for(i=0; i<vector_size(v); i++) {
                    sFile_delete(vector_item(v, i));
                }
                vector_clear(v);

                readline_select_file_read(v, current_dir);
                cursor = 0;
            }

            ///  ///
            if(cursor < 0) {
                cursor = 0;
            }
            if(cursor >= vector_size(v)) {
                cursor = vector_size(v)-1;
            }
        }
    }

    int i;
    for(i=0; i<vector_size(v); i++) {
        sFile_delete(vector_item(v, i));
    }
    vector_delete(v);

    for(i=0; i<vector_size(mark_files); i++) {
        string_delete(vector_item(mark_files, i));
    }
    vector_delete(mark_files);

//mclear();
//mrefresh();
    mendwin();
    //mrestore_screen();

    return 0;
}
*/
