#include "kitutuki.h"
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <locale.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <glob.h>
#include <libgen.h>
#include <dirent.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <oniguruma.h>

#include "config.h"

#if defined(HAVE_CURSES_H)
#include <curses.h>
#elif defined(HAVE_NCURSES_H)
#include <ncurses.h>
#elif defined(HAVE_NCURSES_NCURSES_H)
#include <ncurses/ncurses.h>
#endif

#ifdef __LINUX__
#include <sys/stat.h>
#include <linux/stat.h>
#endif
#if defined(__CYGWIN__) || defined(__CYGWIN7__)
#include <sys/stat.h>
#endif
#
#ifdef __LINUX__
#include <linux/limits.h>
#endif

#ifdef __LINUX__

#include <linux/limits.h>
#define __USE_POSIX
#define __USE_XOPEN_EXTENDED
#define __USE_POSIX199309
#define __USE_UNIX98
#include <signal.h>
#include <sys/wait.h>

#else

#include <signal.h>
#include <sys/stat.h>

#endif

#include "minato_extra.h"

void srandom(unsigned int seed);
char *strdup(const char *s);
int setenv(const char *name, const char *value, int overwrite);
int unsetenv(const char *name);
int strcasecmp(const char *s1, const char *s2);

#define PATH_MAX 4096

/// from cmdline.c ///
char* mreadline(char* prompt);
void madd_history(char* cmdline);
void mwrite_history(char* file);
int mread_history(char* file);
void mremove_history(int num);
void cmdline_init();
void cmdline_final();

/// from readline.c ///
char** readline_on_complete(const char* text, int start, int end);

//////////////////////////////////////////////////////////////////////
// exit時に実行される関数
//////////////////////////////////////////////////////////////////////
static void atexit_fun()
{
    if(mis_raw_mode()) {
        mendwin();
        //mreset_tty();
    }
}


//////////////////////////////////////////////////////////////////////
// メイン関数
//////////////////////////////////////////////////////////////////////
static void usage()
{
    printf("usage kitutuki [-c command] [ -s -w -e ] [ -Lu -Lm -Lw ] [--version ] [ script file]\n\n");
    printf("-c : eval a command on kitutuki\n");
    printf("-rs : use run time a source run time script\n");
    printf("-ro : use run time a compiled run time script(default)\n");
    printf("-rn : no use run time script\n");
    printf("-ts : set termianl kanjicode sjis\n");
    printf("-tw : set termianl kanjicode utf8\n");
    printf("-te : set termianl kanjicode eucjp\n");
    printf("-nr : no use readline\n");
    printf("-s : set kitutuki kanjicode sjis\n");
    printf("-w : set kitutuki kanjicode utf8\n");
    printf("-e : set kitutuki kanjicode eucjp\n");
    printf("-Lu : set line field LF\n");
    printf("-Lm : set line field CR\n");
    printf("-Lw : set line field CRLF\n");
    printf("--version : display kitutuki version\n");

    exit(0);
}

static void version()
{
    printf("kitutuki version %s.", getenv("KITUTUKI_VERSION"));
    printf("author Daisuke Minato.");
    
    exit(0);
}

void main_kitutuki_job_done(int job_num, char* job_title)
{
    printf("%d: %s done.\n", job_num+1, job_title);
}

static int mreadline_signal()
{
    struct sigaction sa2;
    memset(&sa2, 0, sizeof(sa2));
    sa2.sa_handler = SIG_DFL;
//    sa2.sa_flags |= SA_RESTART;
    if(sigaction(SIGINT, &sa2, NULL) < 0) {
        perror("sigaction2");
        exit(1);
    }
}

void sig_int_readline(int signo)
{
    rl_kill_full_line(0,0);
    rl_reset_line_state();
    rl_crlf();
    rl_redisplay();
}

static int readline_signal()
{
    struct sigaction sa2;
    memset(&sa2, 0, sizeof(sa2));
//    sa2.sa_handler = SIG_DFL;
    sa2.sa_handler = sig_int_readline;
    sa2.sa_flags |= SA_RESTART;
    if(sigaction(SIGINT, &sa2, NULL) < 0) {
        perror("sigaction2");
        exit(1);
    }
}

int main(int argc, char* argv[])
{
    /// メモリリークの監視を開始 ///
    CHECKML_BEGIN();

    /// 乱数初期化 ///
    srandom(1000);

    /// ロケール初期化 ///
    setlocale(LC_CTYPE, "ja_JP.UTF-8");

    /// オプション処理 ///
    char send_cmd[BUFSIZ];
    char file_name[PATH_MAX];
    file_name[0] = 0;

    setenv("LINEFIELD","LF", 1);
    enum eTerminalKanjiCode code = kTKUtf8;

    int i;
    enum eRuntimeScript runtime_script = kRSObject;
    BOOL use_readline = TRUE;
    for(i=1 ; i<argc; i++){
        // ファイル名
        if(argv[i][0] != '-') {
            strcpy(file_name, argv[i]);
        }
        // 引数
        else {
            if(strcmp(argv[i], "--version") == 0) {
                version();
                break;
            }
            else if(strcmp(argv[i], "-ts") == 0) {
                code = kTKSjis;
            }
            else if(strcmp(argv[i], "-tw") == 0) {
                code = kTKUtf8;
            }
            else if(strcmp(argv[i], "-te") == 0) {
                code = kTKEucjp;
            }
            else if(strcmp(argv[i], "-rs") == 0) {
                runtime_script = kRSSource;
            }
            else if(strcmp(argv[i], "-ro") == 0) {
                runtime_script = kRSObject;
            }
            else if(strcmp(argv[i], "-rn") == 0) {
                runtime_script = kRSNoRead;
            }
            else if(strcmp(argv[i], "-nr") == 0) {
                use_readline = FALSE;
            }
            else {
                switch (argv[i][1]){
                    case 'c':
                        if(i+1 < argc) {
                            sprintf(send_cmd, "c%s", argv[i+1]);
                            i++;
                        }
                        else {
                            usage();
                        }
                        break;

                    case 's':
                        gKanjiCode = kSjis;
                        break;

                    case 'e':
                        gKanjiCode = kEucjp;
                        break;

                    case 'w':
                        gKanjiCode = kUtf8;
                        break;

                    case 'L':
                        if(strcmp(argv[i], "-Lu") == 0) {
                            gScriptLineField = kLF;
                        }
                        else if(strcmp(argv[i], "-Lw") == 0) {
                            gScriptLineField = kCRLF;
                        }
                        else if(strcmp(argv[i], "-Lm") == 0) {
                            gScriptLineField = kCR;
                        }
                        else {
                            fprintf(stderr, "invalid option %s", argv[i]);
                            exit(1);
                        }
                        break;

                    default:
                        usage();
                }
            }
        }
    }

    /// 引数のファイル名がディレクトリじゃなければスクリプトと判断する ///
    if(strcmp(file_name, "") != 0) {
        struct stat stat_;
        if(stat(file_name, &stat_) < 0) {
            printf("argument file name is invalid\n");
            exit(1);
        }

        if(!S_ISDIR(stat_.st_mode)) {
            mcurses_init(code);
            kitutuki_init(kATConsoleApp, FALSE, runtime_script);
            kitutuki_set_signal();
            int rcode = kitutuki_load(file_name, STDOUT_FILENO, STDIN_FILENO, STDERR_FILENO);
            if(rcode != 0) {
                fprintf(stderr, "run time err\n");
                fprintf(stderr, "%s", gErrMsg);
            }
            kitutuki_final();
            mcurses_final();
            CHECKML_END();

            return rcode;
        }
    }

    /// オプションで指定されているならシェルで実行する ///
    if(send_cmd[0] == 'c') {
        mcurses_init(code);
        kitutuki_init(kATOptC, FALSE, runtime_script);
        kitutuki_set_signal();
        int rcode = kitutuki_shell(send_cmd + 1, "kitutuki -a", STDOUT_FILENO, STDIN_FILENO, STDERR_FILENO);
        if(rcode < 0) { 
            fprintf(stderr, "run time err\n");
            fprintf(stderr, "%s", gErrMsg);
        }
        kitutuki_final();
        mcurses_final();
        CHECKML_END();

        return rcode;
    }

    /// インタラクティブシェル開始 ///
    mcurses_init(code);
    cmdline_init();

    kitutuki_init(kATConsoleApp, TRUE, runtime_script);
    kitutuki_rehash();
    kitutuki_job_done = main_kitutuki_job_done;

    /// ヒストリーを読み込む ///
    char* histfile = getenv("KITUTUKI_HISTFILE");
    if(histfile) {
        if(use_readline) {
            read_history(histfile);
        }
        else {
            mread_history(histfile);
        }
    }
    int histsize = atoi(getenv("KITUTUKI_HISTSIZE"));
    if(histsize < 0) histsize = 1000;

    kitutuki_set_signal();

    if(use_readline) {
        /// readline 初期化 ///
        rl_attempted_completion_function = readline_on_complete;
        rl_completer_quote_characters = "\"'";
        rl_completer_word_break_characters = " \t\n\"'|{}&;()<>=";
        rl_filename_quote_characters = " \t\n\\\"|$%*?[]<>:{}&;#()'";
        rl_completion_append_character= ' ';

        char* buf;
        int history_num = 0;
        while(1) {
            /// 一行読み込み ///
            readline_signal();

            /// 終了コードの保存 ///
            char end_code[128];
            char* lvar =  kitutuki_get_local_var("RCODE");
            if(lvar) 
                sprintf(end_code, "%s", lvar);
            else 
                sprintf(end_code, "0");

            /// プロンプト ///
            char* prompt = getenv("KITUTUKI_PROMPT");
            if(prompt) {
                string_obj* ret = STRING_NEW("");
                if(kitutuki_shell3(ret, prompt, "KITUTUKI_PROMPT") == -1) {
                    buf = readline("> ");
                }
                else {
                    buf = readline(string_c_str(ret));
                }

                string_delete(ret);
            }
            else {
                buf = readline("> ");
            }

            /// 終了コードの復帰 ///
            kitutuki_set_local_var("RCODE", end_code);

            /// 実行 ///
            if(buf == NULL) {
                break;
            }
            else if(*buf) {
                kitutuki_set_signal();
                int rcode = kitutuki_shell(buf, "kitutuki", STDOUT_FILENO, STDIN_FILENO, STDERR_FILENO);
                if(rcode < 0) {
                    /// ヒストリに追加 ///
                    add_history(buf);
                    history_num++;
                    if(history_num > histsize) {
                        HIST_ENTRY* history = remove_history(0);
                        free(history);
                    }

                    fprintf(stderr, "run time err\n");
                    fprintf(stderr, "%s", gErrMsg);
                }
                else {
                    /// ヒストリに追加 ///
                    add_history(buf);
                    history_num++;
                    if(history_num > histsize) {
                        HIST_ENTRY* history = remove_history(0);
                        free(history);
                    }
                }
            }

            /// バックグランド待ち ///
            kitutuki_wait_background_job();

            free(buf);

            if(gSaphireExit >= 0) break;        // exitコマンドが実行された
        }
    }
    else {
        char* buf;
        int history_num = 0;
        while(1) {
            /// 一行読み込み ///
            mreadline_signal();

            /// 終了コードの保存 ///
            char end_code[128];
            char* lvar =  kitutuki_get_local_var("RCODE");
            if(lvar) 
                sprintf(end_code, "%s", lvar);
            else 
                sprintf(end_code, "0");

            /// プロンプト ///
            char* prompt = getenv("KITUTUKI_PROMPT");
            if(prompt) {
                string_obj* ret = STRING_NEW("");
                if(kitutuki_shell3(ret, prompt, "KITUTUKI_PROMPT") == -1) {
                    buf = mreadline("> ");
                }
                else {
                    buf = mreadline(string_c_str(ret));
                }

                string_delete(ret);
            }
            else {
                buf = mreadline("> ");
            }

            /// 終了コードの復帰 ///
            kitutuki_set_local_var("RCODE", end_code);

            /// 実行 ///
            if(buf == NULL) {
                break;
            }
            else if(*buf) {
                kitutuki_set_signal();
                int rcode = kitutuki_shell(buf, "kitutuki", STDOUT_FILENO, STDIN_FILENO, STDERR_FILENO);
                if(rcode == SIGSEGV) {
                    /// ヒストリに追加 ///
                    madd_history(buf);
                    history_num++;
                    if(history_num > histsize) {
                        mremove_history(0);
                    }

                    fprintf(stderr, "SEGMENTATION FAULT\n");
                }
                else if(rcode < 0) {
                    /// ヒストリに追加 ///
                    madd_history(buf);
                    history_num++;
                    if(history_num > histsize) {
                        mremove_history(0);
                    }

                    fprintf(stderr, "run time err\n");
                    fprintf(stderr, "%s", gErrMsg);
                }
                else {
                    /// ヒストリに追加 ///
                    madd_history(buf);
                    history_num++;
                    if(history_num > histsize) {
                        mremove_history(0);
                    }
                }
            }

            /// バックグランド待ち ///
            kitutuki_wait_background_job();

            if(gSaphireExit >= 0) break;        // exitコマンドが実行された
        }
    }

    kitutuki_kill_all_jobs();
    kitutuki_final();
    mcurses_final();
    cmdline_final();

    /// ヒストリーを書き込む ///
    if(histfile) {
        if(use_readline) {
            write_history(histfile);
        }
        else {
            mwrite_history(histfile);
        }
    }
    
    mreset_tty();

    /// メモリリーク検出 ///
    CHECKML_END();

    return 0;
}

