#define _XOPEN_SOURCE 500

#include "config.h"
#include "minato_debug.h"
#include "minato_kanji.h"
#include "minato_curses.h"
#include <wctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>


char* gKanjiCodeString[5] = {
    "eucjp", "sjis", "utf8", "utf8-mac", "unknown"
};

char* gIconvKanjiCodeName[5] = {
    "eucjp", "sjis", "utf-8", "utf-8-mac", "unknown"
};

#if defined(__CYGWIN__) || defined(__CYGWIN7__)
char 	*_EXFUN(strdup,(const char *));
#endif

//////////////////////////////////////////////////////////////////////
// 初期化
//////////////////////////////////////////////////////////////////////
void kanji_init()
{
}

//////////////////////////////////////////////////////////////////////
// 解放
//////////////////////////////////////////////////////////////////////
void kanji_final()
{
}

//////////////////////////////////////////////////////////////////////
// 引数の文字列が画面で何文字か(範囲指定)
//////////////////////////////////////////////////////////////////////
int str_termlen_range(enum eKanjiCode code, char* mbs, int utfpos1, int utfpos2)
{
    if(code == kUtf8) {
        if(is_utf8_bytes(mbs) && utfpos1 < utfpos2) {
            char* putfpos1 = str_kanjipos2pointer(kUtf8, mbs, utfpos1);
            char* putfpos2 = str_kanjipos2pointer(kUtf8, mbs, utfpos2);
            
            char* mbs2 = MALLOC(putfpos2 - putfpos1 + 1);
            memcpy(mbs2, putfpos1, putfpos2 - putfpos1);
            mbs2[putfpos2 - putfpos1] = 0;

            int result = str_termlen(kUtf8, mbs2);
            FREE(mbs2);

            return result;
        }
        else {
            return -1;
        }
    }
    else {
        fprintf(stderr, "not defined kEucjp, kSjis");
        exit(1);
    }
}

//////////////////////////////////////////////////////////////////////
// 文字列がUTF8文字列かどうか
//////////////////////////////////////////////////////////////////////
BOOL is_utf8_bytes(char* mbs)
{
    const int size = (strlen(mbs)+1)*MB_CUR_MAX;
    wchar_t* wcs = (wchar_t*)MALLOC(size);
    char* mbs2 = MALLOC(size);

    if(mbstowcs(wcs, mbs, size) < 0) {
        FREE(wcs);
        FREE(mbs2);
        return FALSE;
    }

    if(wcstombs(mbs2, wcs, size) < 0) {
        FREE(wcs);
        FREE(mbs2);
        return FALSE;
    }

    if(strcmp(mbs, mbs2) == 0) {
        FREE(wcs);
        FREE(mbs2);
        return TRUE;
    }
    else {
        FREE(wcs);
        FREE(mbs2);
        return FALSE;
    }
}


//////////////////////////////////////////////////////////////////////
// mbsを漢字文字列とみたときのpointの場所は何文字目かを返す
//////////////////////////////////////////////////////////////////////
int str_pointer2kanjipos(enum eKanjiCode code, char* mbs, char* point)
{
    if(code == kEucjp || code == kSjis) {
        int result = 0;

        char* p = mbs;
        while(p < point) {
            if(is_kanji(code, *p)) {
                p+=2;
                result++;
            }
            else {
                p++;
                result++;
            }
        }

        return result;
    }
    else {
        char* mbs2;
        int result;

        mbs2 = STRDUP(mbs);
        *(mbs2 + (point-mbs)) = 0;

        result = str_kanjilen(code, mbs2);

        FREE(mbs2);

        return result;
    }
}

//////////////////////////////////////////////////////////////////////
// mbsをsjis,eucjp,UTF文字列と見たときのpos文字目の位置を返す
//////////////////////////////////////////////////////////////////////
char* str_kanjipos2pointer(enum eKanjiCode code, char* mbs, int pos)
{
    if(code == kUtf8) {
        char* result;
        wchar_t* wcs;
        char* mbs2;
        
        if(pos<0) return mbs;
        if(pos >= str_kanjilen(code, mbs)) return mbs + strlen(mbs);

        mbs2 = (char*)MALLOC(sizeof(char)*(strlen(mbs)+1));
        wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*(strlen(mbs)+1)*MB_CUR_MAX);

        if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) == -1) {
            mbstowcs(wcs, "?????", (strlen(mbs)+1)*MB_CUR_MAX);
        }

        wcs[pos] = 0;

        wcstombs(mbs2, wcs, strlen(mbs)+1);

        result = mbs + strlen(mbs2);

        FREE(wcs);
        FREE(mbs2);

        return result;
    }
    else {
        if(pos<0) return mbs;
        if(pos >= str_kanjilen(code, mbs)) return mbs + strlen(mbs);

        char* p = mbs;
        int c = 0;
        while(*p) {
            if(c==pos) {
                return p;
            }

            if(is_kanji(code, *p)) {
                p+=2;
                c++;
            }
            else {
                p++;
                c++;
            }
        }

        return NULL;
    }
}

//////////////////////////////////////////////////////////////////////
// 引数が漢字で何文字か
//////////////////////////////////////////////////////////////////////
int str_kanjilen(enum eKanjiCode code, char* mbs)
{
    if(code == kUtf8) {
        int result;
        wchar_t* wcs;
        
        wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*(strlen(mbs)+1)*MB_CUR_MAX);

        if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) == -1) {
            FREE(wcs);
            return -1;
        }

        result = wcslen(wcs);

        FREE(wcs);

        return result;
    }
    else {
        int result = 0;
        char* p = mbs;
        while(*p) {
            if(is_kanji(code, *p) && *(p+1) != 0) {
                p+=2;
                result++;
            }
            else {
                p++;
                result++;
            }
        }
        return result;
    }
}


/*
//////////////////////////////////////////////////////////////////////
// mbsを漢字文字列とみたときのpointの場所は何文字目かを返す
//////////////////////////////////////////////////////////////////////
int str_pointer2kanjipos(enum eKanjiCode code, char* mbs, char* point)
{
    if(code == kEucjp || code == kSjis) {
        int result = 0;

        char* p = mbs;
        while(p < point) {
            if(is_kanji(code, *p)) {
                p+=2;
                result++;
            }
            else {
                p++;
                result++;
            }
        }

        return result;
    }
    else {
        char* mbs2;
        int result;

        mbs2 = STRDUP(mbs);
        *(mbs2 + (point-mbs)) = 0;

        result = str_kanjilen(code, mbs2);

        FREE(mbs2);

        return result;
    }
}

//////////////////////////////////////////////////////////////////////
// mbsをsjis,eucjp,UTF文字列と見たときのpos文字目の位置を返す
//////////////////////////////////////////////////////////////////////
char* str_kanjipos2pointer(enum eKanjiCode code, char* mbs, int pos)
{
    if(code == kUtf8) {
        if(!is_utf8_bytes(mbs) || pos < 0 || pos >= str_kanjilen(code,mbs))
        {
            return NULL;
        }

        char* result;
        wchar_t* wcs;
        char* mbs2;

        mbs2 = (char*)MALLOC(sizeof(char)*(strlen(mbs)+1));
        wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*(strlen(mbs)+1)*MB_CUR_MAX);

        if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) < 0) {
            FREE(mbs2);
            FREE(wcs);
            return NULL;
        }

        wcs[pos] = 0;

        if(wcstombs(mbs2, wcs, strlen(mbs)+1) < 0) {
            FREE(mbs2);
            FREE(wcs);
            return NULL;
        }

        result = mbs + strlen(mbs2);

        FREE(wcs);
        FREE(mbs2);

        return result;
    }
    else {
        if(pos<0) return mbs;
        if(pos >= str_kanjilen(code, mbs)) return mbs + strlen(mbs);

        char* p = mbs;
        int c = 0;
        while(*p) {
            if(c==pos) {
                return p;
            }

            if(is_kanji(code, *p)) {
                p+=2;
                c++;
            }
            else {
                p++;
                c++;
            }
        }

        return NULL;
    }
}

//////////////////////////////////////////////////////////////////////
// 引数が漢字で何文字か
//////////////////////////////////////////////////////////////////////
int str_kanjilen(enum eKanjiCode code, char* mbs)
{
    if(code == kUtf8) {
        if(!is_utf8_bytes(mbs)) {
            return -1;
        }
        else {
            int result;
            wchar_t* wcs;
            
            wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*(strlen(mbs)+1)*MB_CUR_MAX);

            if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) < 0) {
                FREE(wcs);
                return -1;
            }

            result = wcslen(wcs);

            FREE(wcs);

            return result;
        }
    }
    else {
        int result = 0;
        char* p = mbs;
        while(*p) {
            if(is_kanji(code, *p) && *(p+1) != 0) {
                p+=2;
                result++;
            }
            else {
                p++;
                result++;
            }
        }
        return result;
    }
}
*/















//////////////////////////////////////////////////////////////////////
// 半角文字一文字目かどうか
//////////////////////////////////////////////////////////////////////
int is_hankaku(enum eKanjiCode code, unsigned char c)
{
    if(code == kEucjp)
        return c == 0x8e;
    else
        return 0;
}

//////////////////////////////////////////////////////////////////////
// 漢字一文字目かどうか
//////////////////////////////////////////////////////////////////////
int is_kanji(enum eKanjiCode code, unsigned char c)
{
    if(code == kEucjp) {
        return c >= 0xA1 && c <= 0xFE || c == 0x8e;
//        return c >= 161 && c <= 254;
    }
    else if(code == kSjis) {
        return c >= 0x81 && c <= 0x9f || c >= 0xE0;
    }
    else if(code == kUtf8) {
        return !(c>=' ' && c<='~');
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////
// FALSE:漢字まじり TRUE:全部英数字
//////////////////////////////////////////////////////////////////////
int is_all_ascii(char* buf)
{
    char* p = buf;
    while(*p) {
        if(*p < ' ' || *p > '~') {
            return FALSE;
        }
        p++;
    }

    return TRUE;
}

//////////////////////////////////////////////////////////////////////
// 引数の文字列が画面で何文字か
//////////////////////////////////////////////////////////////////////
int str_termlen(enum eKanjiCode code, char* mbs)
{
    if(code == kUtf8) {
        int result;
        wchar_t* wcs;

        wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*((strlen(mbs)+1)*MB_CUR_MAX));
        if(mbstowcs(wcs, mbs, (strlen(mbs)+1)*MB_CUR_MAX) == -1) {
            FREE(wcs);
            return -1;
        }
        result = wcswidth(wcs, wcslen(wcs));
        FREE(wcs);

        return result;
    }
    else {
        return strlen(mbs);
    }
}

//////////////////////////////////////////////////////////////////////
// 漢字でpos文字目までの端末で何文字か
//////////////////////////////////////////////////////////////////////
int str_termlen2(enum eKanjiCode code, char* mbs, int pos)
{
    if(code == kUtf8) {
        char* mbs2;
        int result;
        char* point;

        mbs2 = STRDUP(mbs);
        point = str_kanjipos2pointer(kUtf8, mbs2, pos);
        if(point == NULL) {
            if(pos < 0) 
                result = -1;
            else
                result = str_termlen(code, mbs2);
        }
        else {
            *(mbs2 + (point-mbs2)) = 0;
            result = str_termlen(code, mbs2);
        }

        FREE(mbs2);

        return result;
    }
    else {
        return pos*2;
    }
}


//////////////////////////////////////////////////////////////////////
// mbsから端末上の文字列分(termsize)のみ残して残りを捨てた文字列を
// dest_mbsに返す dest_byte: dest_mbsのサイズ
//////////////////////////////////////////////////////////////////////
void str_cut(enum eKanjiCode code, char* mbs, int termsize, SPACE char* dest_mbs, int dest_byte)
{
    if(code == kUtf8) {
        wchar_t* wcs;
        wchar_t* wcs_tmp;
        int i;

        wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*(termsize+1)*MB_CUR_MAX);
        wcs_tmp = (wchar_t*)MALLOC(sizeof(wchar_t)*(termsize+1)*MB_CUR_MAX);
        if(mbstowcs(wcs, mbs, (termsize+1)*MB_CUR_MAX) == -1) {
            mbstowcs(wcs, "?????", (termsize+1)*MB_CUR_MAX);
        }

        for(i=0; i<wcslen(wcs); i++) {
            wcs_tmp[i] = wcs[i];
            wcs_tmp[i+1] = 0;

            if(wcswidth(wcs_tmp, wcslen(wcs_tmp)) > termsize) {
                wcs_tmp[i] = 0;
                break;
            }
        }

        wcstombs(dest_mbs, wcs_tmp, dest_byte);

        FREE(wcs);
        FREE(wcs_tmp);
    }
    else {
        int n;
        BOOL kanji = FALSE;
        for(n=0; n<termsize && n<dest_byte-1; n++) {
            if(!kanji && is_kanji(code, mbs[n])) {
                kanji = TRUE;
            }
            else {
                kanji = FALSE;
            }

            dest_mbs[n] = mbs[n];
        }
        
        if(kanji)
            dest_mbs[n-1] = 0;
        else
            dest_mbs[n] = 0;
    }
}
//////////////////////////////////////////////////////////////////////
// mbsから端末上の文字列分(termsize)のみ残して残りを捨てた文字列を
// dest_mbsに返す dest_byte: dest_mbsのサイズ
// 切り捨てた文字列をスペースで埋める
//////////////////////////////////////////////////////////////////////
#include <stdio.h>

void str_cut2(enum eKanjiCode code, char* mbs, int termsize, SPACE char* dest_mbs, int dest_byte)
{
    if(code == kUtf8) {
        int i;
        int n;

        wchar_t* wcs 
            = (wchar_t*)MALLOC(sizeof(wchar_t)*(termsize+1)*MB_CUR_MAX);
        wchar_t* tmp 
            = (wchar_t*)MALLOC(sizeof(wchar_t)*(termsize+1)*MB_CUR_MAX);

        if(mbstowcs(wcs, mbs, (termsize+1)*MB_CUR_MAX) == -1) {
            mbstowcs(wcs, "?????", (termsize+1)*MB_CUR_MAX);
        }

        i = 0;
        while(1) {
            if(i < wcslen(wcs)) {
                tmp[i] = wcs[i];
                tmp[i+1] = 0;
            }
            else {
                tmp[i] = ' ';
                tmp[i+1] = 0;
            }

            n = wcswidth(tmp, wcslen(tmp));
            if(n < 0) {
                strcpy(dest_mbs, "?????");

                FREE(wcs);
                FREE(tmp);

                return;
            }
            else {
                if(n > termsize) {
                    tmp[i] = 0;

                    if(wcswidth(tmp, wcslen(tmp)) != termsize) {
                        tmp[i] = ' ';
                        tmp[i+1] = 0;
                    }
                    break;
                }
            }

            i++;
        }

        wcstombs(dest_mbs, tmp, dest_byte);

        FREE(wcs);
        FREE(tmp);
    }
    else {
        int n;
        BOOL tmpend = FALSE;
        BOOL kanji = FALSE;
        for(n=0; n<termsize && n<dest_byte-1; n++) {
            if(kanji)
                kanji = FALSE;
            else if(!tmpend && is_kanji(code, mbs[n])) {
                kanji = TRUE;
            }

            if(!tmpend && mbs[n] == 0) tmpend = TRUE;
                        
            if(tmpend)
                dest_mbs[n] = ' ';
            else
                dest_mbs[n] = mbs[n];
        }
        
        if(kanji) dest_mbs[n-1] = ' ';
        dest_mbs[n] = 0;
    }
}

//////////////////////////////////////////////////////////////////////
// mbsから端末上の文字列分(termsize)のみ残して残りを捨てた文字列を
// dest_mbsに返す 
// dest_byte: dest_mbsのサイズ
// 切り捨てた文字列をスペースで埋める
// 切り捨てるのは文字列の前から
//////////////////////////////////////////////////////////////////////
void str_cut3(enum eKanjiCode code, char* mbs, int termsize, SPACE char* dest_mbs, int dest_byte)
{
    if(code == kUtf8) {
        wchar_t* wcs;
        wchar_t* tmp;
        int i;
        int j;

        wcs = (wchar_t*)MALLOC(sizeof(wchar_t)*(strlen(mbs)+1)*MB_CUR_MAX);
        tmp = (wchar_t*)MALLOC(sizeof(wchar_t)*(strlen(mbs)+1)*MB_CUR_MAX);

        if(str_termlen(code, mbs) <= termsize) {
            str_cut2(code, mbs, termsize, dest_mbs, dest_byte);
        }
        else {
            if(mbstowcs(wcs, mbs, (termsize+1)*MB_CUR_MAX) == -1) {
                mbstowcs(wcs, "?????", (termsize+1)*MB_CUR_MAX);
            }

            i = wcslen(wcs)-1;
            j = 0;
            while(1) {
                tmp[j++] = wcs[i--];
                tmp[j] = 0;

                if(wcswidth(tmp, wcslen(tmp)) > termsize) {
                    tmp[j-1] = 0;
                    i+=2;

                    if(wcswidth(tmp, wcslen(tmp)) == termsize) {
                        wcscpy(tmp, wcs + i);
                        break;
                    }
                    else {
                        tmp[0] = ' ';
                        wcscpy(tmp + 1, wcs + i);
                        break;
                    }
                }
            }

            wcstombs(dest_mbs, tmp, dest_byte);
        }

        FREE(wcs);
        FREE(tmp);
    }
    else {
        BOOL* kanji;
        int i;
        const int len = strlen(mbs);
        char* buf;

        if(len < termsize) {
            strcpy(dest_mbs, mbs);
            for(i=len; i<termsize; i++) {
                strcat(dest_mbs, " ");
            }
        }
        else {
            kanji = MALLOC(sizeof(BOOL)*len);
            for(i=0; i<len; i++) {
                if(is_kanji(code, mbs[i])) {
                    kanji[i] = TRUE;
                    i++;
                    kanji[i] = FALSE;
                }
                else {
                    kanji[i] = FALSE;
                }
            }

            if(kanji[len-termsize-1]) {
                int i;
                for(i=len-termsize; i<len; i++) {
                    dest_mbs[i-len+termsize] = mbs[i];
                }
                dest_mbs[i-len+termsize] = 0;
                dest_mbs[0] = ' ';
            }
            else {
                int i;
                for(i=len-termsize; i<len; i++) {
                    dest_mbs[i-len+termsize] = mbs[i];
                }
                dest_mbs[i-len+termsize] = 0;
            }

            FREE(kanji);
        }
    }
}

//////////////////////////////////////////////////////////////////////
// 漢字エンコードの変換。iconvのラッパー
//////////////////////////////////////////////////////////////////////
#if defined(HAVE_ICONV_H) && !defined(__CYGWIN7__) && !defined(__CYGWIN__)
#include <iconv.h>

extern size_t iconv (iconv_t cd,  char* * inbuf, size_t *inbytesleft, char* * outbuf, size_t *outbytesleft);

int kanji_convert(char* input_buf, char* output_buf, size_t output_buf_size, enum eKanjiCode input_kanji_encode_type, enum eKanjiCode output_kanji_encode_type)
{
    iconv_t ic;
    char* ptr_in;
    char* ptr_out;
    const int len = strlen(input_buf);
    size_t result;
    size_t bufsz_in;
    size_t bufsz_out;

    if(is_all_ascii(input_buf)
        || input_kanji_encode_type == output_kanji_encode_type
        || input_kanji_encode_type < 0
        || input_kanji_encode_type >= kUnknown 
        || output_kanji_encode_type < 0
        || output_kanji_encode_type >= kUnknown)
    {
        strcpy(output_buf, input_buf);
        return 0;
    }
    
    ptr_in = input_buf;
    ptr_out = output_buf;
    bufsz_in = (size_t) len;
    bufsz_out = (size_t) output_buf_size;

    ic = iconv_open(gIconvKanjiCodeName[output_kanji_encode_type]
                    , gIconvKanjiCodeName[input_kanji_encode_type]);

    if(iconv(ic, &ptr_in, &bufsz_in, &ptr_out, &bufsz_out) < 0 
            || len-(int)bufsz_in < len) 
    {
        iconv_close(ic);

        return -1;
    }
    output_buf[output_buf_size - bufsz_out] = 0;

    iconv_close(ic);

    return 0;
}



#endif
