#include "kitutuki.h"
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <locale.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <glob.h>
#include <libgen.h>
#include <dirent.h>
#include <readline/readline.h>
#include <readline/history.h>
#include <oniguruma.h>

#include "config.h"

#if defined(HAVE_CURSES_H)
#include <curses.h>
#elif defined(HAVE_NCURSES_H)
#include <ncurses.h>
#elif defined(HAVE_NCURSES_NCURSES_H)
#include <ncurses/ncurses.h>
#endif

#ifdef __LINUX__
#include <sys/stat.h>
#include <linux/stat.h>
#endif
#if defined(__CYGWIN__) || defined(__CYGWIN7__)
#include <sys/stat.h>
#endif
#
#ifdef __LINUX__
#include <linux/limits.h>
#endif

#ifdef __LINUX__

#include <linux/limits.h>
#define __USE_POSIX
#define __USE_XOPEN_EXTENDED
#define __USE_POSIX199309
#define __USE_UNIX98
#include <signal.h>
#include <sys/wait.h>

#else

#include <signal.h>
#include <sys/stat.h>

#endif

#include "minato_extra.h"

void srandom(unsigned int seed);
char *strdup(const char *s);
int setenv(const char *name, const char *value, int overwrite);
int unsetenv(const char *name);
int strcasecmp(const char *s1, const char *s2);

#define PATH_MAX 4096

//////////////////////////////////////////////////////////////////////
// readline
//////////////////////////////////////////////////////////////////////
char* readline_program_completion(const char* text, int stat)
{
    static int index, wordlen;

    if(stat == 0) {
        wordlen = strlen(text);
        index = 0;
    }

    while(index < vector_size(gPrograms)) {
        char* program = string_c_str(vector_item(gPrograms, index));
        index++;

        if(!strncmp(text, program, wordlen)) {
            return strdup(program);
        }
    }

    return NULL;
}

char** readline_on_complete(const char* text, int start, int end)
{
//printf("\n\ntext (%s) start %d end %d rl_line_buffer %s\n\n", text, start, end, rl_line_buffer);
    
    BOOL program = TRUE;

    char* p = rl_line_buffer;

    while(*p == ' ' || *p == '\t' || *p == '\n') {
        p++;
    }

    BOOL squote = FALSE;
    BOOL dquote = FALSE;
    while(p < p + start) {
        // クォート
        if(*p == '\\' && *(p+1) == 't') {
            p+=2;
        }
        else if(*p == '\\' && *(p+1) == 'n') {
            p+=2;
        }
        else if(*p == '\\' && *(p+1) == 'r') {
            p+=2;
        }
        else if(*p == '\\') {
            p+=2;
        }

        // シングルクォート
        else if(!dquote && *p == '\'') {
            p++;
            squote = !squote;
        }
        // ダブルクォート
        else if(!squote && *p == '"') {
            p++;
            dquote = !dquote;
        }
        // 環境変数 $$()
        else if(!squote && *p == '$' && *(p+1) == '$' && *(p+2) == '(') {
            p+=3;

            program = TRUE;

            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        // 環境変数 $()
        else if(!squote && *p == '$' && *(p+1) == '(') {
            p+=2;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        else if(!dquote && !squote && *p == ')') {
            p++;
            program = FALSE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }

        // 環境変数 $$
        else if(!squote && *p == '$' && *(p+1) == '$') {
            p+=2;
        }
        // 環境変数 $
        else if(!squote && *p == '$') {
            p+=2;
        }
        /// シングルクォート、ダブルクォート中 ///
        else if(squote || dquote) {
            p++;
        }
        /// グロブ ///
        else if(*p == '*' || *p == '?' || *p =='[' && (*(p+1) != ' ' && *(p+1) != '\t' && *(p+1) == '\n')) 
        {
            p++;
        }
        else if(*p == '<' && program) {
            p++;

            program = FALSE;
        }
/*
        /// ファイルハンドル1 ///
        else if(strcmp(string_c_str(buf), "") ==0 
        }
                && parse_is_file_handle(p)) 
        /// ファイルハンドル2 ///
        else if(strcmp(string_c_str(buf), "") ==0 
                && parse_is_file_handle2(p))
*/
        /// 単語の区切り /////////////////////////////////////////////////
    
        /// 空白 ///
        else if(*p == ' ' || *p == '\t') {
            while(*p == ' ' || *p == '\t') {
                p++;
            }

            if(program) program = FALSE;
        }
/*
        /// リダイレクション - エラー出力を標準出力にする ///
        else if(parse_is_redirect2(p)) {
        }

        /// リダイレクション ///
        else if(parse_is_redirect(p)) 
        {
        }
*/
        /// ブロック
        else if(*p == ':' || *p == '{') {
            p++;
            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        
        /// コマンドの区切り ///////////////////////////////////////////
        
        /// パイプ ///
        else if(*p == '|' && *(p+1) == '~') {
            p+=2;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        
        /// パイプ ///
        else if(*p == '|' && *(p+1) != '|') {
            p++;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        
        /// 文(statment)の区切り系 ///////////////////////////////////
    
        /// アンドアンド ///
        else if(*p == '&' && *(p+1) == '&') {
            p+=2;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        /// オアオア ///
        else if(*p == '|' && *(p+1) == '|') {
            p+=2;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        /// バックグラウンド ///
        else if(*p == '&' && *(p+1) != '&') {
            p++;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        /// セミコロン ///
        else if(*p == ';') {
            p++;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        /// 改行 ///
        else if(*p == '\n') {
            p++;

            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        /// コメント ///
        else if(*p == '#') {
            p++;

            while(*p) {
                if(*p == '\n') {
                    p++;
                    program = TRUE;
                    break;
                }
                else {
                    p++;
                }
            }
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        /// 終了 ///
        else if(*p == 0) {
            break;
        }
        
        /// 構文 ////////////////////////////////////////////////////////

        /// サブシェル ///
        else if(*p == '(') {
            p++;
            program = TRUE;
            while(*p == ' ' || *p == '\t' || *p == '\n') {
                p++;
            }
        }
        else {
            p++;
        }
    }

    if(program) {
        rl_completion_append_character= ' ';
        return rl_completion_matches(text, readline_program_completion);
    }

    return NULL;
}

